@@ -2,7 +2,7 @@
  * Modified from industrialio buffer test code, and Lenovo Yoga (2 Pro) orientation helper
  * Copyright (c) 2008 Jonathan Cameron
  * Copyright (c) 2014 Peter F. Patel-Schneider
- * Copyright (c) 2014 Bastien Nocera <hadess@hadess.net>
+ * Copyright (c) 2014, 2021 Bastien Nocera <hadess@hadess.net>
  * Copyright (c) 2015 Elad Alfassa <elad@fedoraproject.org>
  *
  * This program is free software; you can redistribute it and/or modify it
@@ -11,6 +11,7 @@
  */
 
 #include "iio-buffer-utils.h"
+#include "utils.h"
 
 #include <fcntl.h>
 #include <string.h>
@@ -54,7 +55,7 @@ struct iio_channel_info {
 static char *
 iioutils_break_up_name (const char *name)
 {
-	char **items, *ret;
+	g_auto(GStrv) items = NULL;
 	guint i;
 
 	items = g_strsplit (name, "_", -1);
@@ -65,10 +66,7 @@ iioutils_break_up_name (const char *name)
 		}
 	}
 
-	ret = g_strjoinv ("_", items);
-	g_strfreev (items);
-
-	return ret;
+	return g_strjoinv ("_", items);
 }
 
 /**
@@ -94,27 +92,25 @@ iioutils_get_type (unsigned   *is_signed,
 		   const char *generic_name)
 {
 	int ret;
-	char *builtname;
-	char *filename;
+	g_autofree char *builtname = NULL;
+	g_autofree char *filename = NULL;
 	char signchar, endianchar;
 	unsigned padint;
-	FILE *sysfsfp;
+	g_autoptr(FILE) sysfsfp = NULL;
 
 	builtname = g_strdup_printf ("%s_type", name);
 	filename = g_build_filename (device_dir, "scan_elements", builtname, NULL);
-	g_free (builtname);
 
 	sysfsfp = fopen (filename, "r");
 	if (sysfsfp == NULL) {
+		g_clear_pointer (&builtname, g_free);
+		g_clear_pointer (&filename, g_free);
 		builtname = g_strdup_printf ("%s_type", generic_name);
 		filename = g_build_filename (device_dir, "scan_elements", builtname, NULL);
-		g_free (builtname);
 
 		sysfsfp = fopen (filename, "r");
-		if (sysfsfp == NULL) {
-			g_free (filename);
+		if (sysfsfp == NULL)
 			return FALSE;
-		}
 	}
 
 	/* See `iio_show_fixed_type()` in the IIO core for the format */
@@ -127,11 +123,8 @@ iioutils_get_type (unsigned   *is_signed,
 
 	if (ret < 0 || ret != 5) {
 		g_warning ("Failed to pass scan type description for %s", filename);
-		fclose (sysfsfp);
-		g_free (filename);
 		return FALSE;
 	}
-	fclose (sysfsfp);
 
 	*be = (endianchar == 'b');
 	*bytes = padint / 8;
@@ -144,8 +137,6 @@ iioutils_get_type (unsigned   *is_signed,
 	g_debug ("Got type for %s: is signed: %d, bytes: %d, bits_used: %d, shift: %d, mask: 0x%" G_GUINT64_FORMAT ", be: %d",
 		 name, *is_signed, *bytes, *bits_used, *shift, *mask, *be);
 
-	g_free (filename);
-
 	return TRUE;
 }
 
@@ -156,53 +147,53 @@ iioutils_get_param_float (float      *output,
 			  const char *name,
 			  const char *generic_name)
 {
-	FILE *sysfsfp;
-	char *builtname, *filename;
-	int ret = 0;
+	g_autofree char *builtname = NULL;
+	g_autofree char *filename = NULL;
+	g_autofree char *contents = NULL;
+	g_autoptr(GError) error = NULL;
 
 	g_debug ("Trying to read '%s_%s' (name) from dir '%s'", name, param_name, device_dir);
 
 	builtname = g_strdup_printf ("%s_%s", name, param_name);
 	filename = g_build_filename (device_dir, builtname, NULL);
-	g_free (builtname);
 
-	sysfsfp = fopen (filename, "r");
-	if (sysfsfp) {
-		ret = fscanf (sysfsfp, "%f", output);
-		fclose (sysfsfp);
-		g_free (filename);
-		if (ret == 1)
+	if (g_file_get_contents (filename, &contents, NULL, &error)) {
+		char *endptr;
+		*output = g_ascii_strtod (contents, &endptr);
+		if (*output != 0.0 || endptr != contents)
 			return 0;
+		g_warning ("Couldn't convert '%s' from %s to float", g_strchomp (contents), filename);
+	} else {
+		g_debug ("Failed to read float from %s: %s", filename, error->message);
+		g_clear_error (&error);
 	}
 
-	ret = -errno;
-	g_debug ("Failed to read float from %s: %s", filename, g_strerror (-ret));
-	g_free (filename);
-
 	g_debug ("Trying to read '%s_%s' (generic name) from dir '%s'", generic_name, param_name, device_dir);
 
+	g_clear_pointer (&builtname, g_free);
+	g_clear_pointer (&filename, g_free);
 	builtname = g_strdup_printf ("%s_%s", generic_name, param_name);
 	filename = g_build_filename (device_dir, builtname, NULL);
-	g_free (builtname);
 
-	sysfsfp = fopen (filename, "r");
-	if (sysfsfp) {
-		if (fscanf (sysfsfp, "%f", output) != 1) {
-			g_debug ("Failed to read float from %s", filename);
-			ret = -EINVAL;
+	g_clear_pointer (&contents, g_free);
+	if (g_file_get_contents (filename, &contents, NULL, &error)) {
+		char *endptr;
+		*output = g_ascii_strtod (contents, &endptr);
+		if (*output == 0.0 && endptr == contents) {
+			g_warning ("Couldn't convert '%s' from %s to float", g_strchomp (contents), filename);
+			return -EINVAL;
 		}
-		fclose (sysfsfp);
 	} else {
-		ret = -errno;
-		if (ret != -ENOENT)
-			g_warning ("Failed to read float from %s: %s", filename, g_strerror (-ret));
-		else
-			g_debug ("Failed to read float from %s: %s", filename, g_strerror (-ret));
+		if (g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
+			g_debug ("Failed to read float from non-existent %s", filename);
+			return -ENOENT;
+		} else {
+			g_warning ("Failed to read float from %s: %s", filename, error->message);
+			return -EINVAL;
+		}
 	}
 
-	g_free (filename);
-
-	return ret;
+	return 0;
 }
 
 static void
@@ -222,13 +213,14 @@ compare_channel_index (gconstpointer a, gconstpointer b)
 	return (int) (info_1->index - info_2->index);
 }
 
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(iio_channel_info, channel_info_free)
+
 /* build_channel_array() - function to figure out what channels are present */
 static iio_channel_info **
 build_channel_array (const char        *device_dir,
 		     int               *counter)
 {
 	GDir *dp;
-	FILE *sysfsfp;
 	int ret;
 	const char *name;
 	char *scan_el_dir;
@@ -251,24 +243,22 @@ build_channel_array (const char        *device_dir,
 
 	while ((name = g_dir_read_name (dp)) != NULL) {
 		if (g_str_has_suffix (name, "_en")) {
-			char *filename, *index_name;
-			iio_channel_info *current;
+			g_autoptr(FILE) sysfsfp = NULL;
+			g_autofree char *filename = NULL;
+			g_autofree char *index_name = NULL;
+			g_autoptr(iio_channel_info) current = NULL;
 
 			filename = g_build_filename (scan_el_dir, name, NULL);
 			sysfsfp = fopen (filename, "r");
 			if (sysfsfp == NULL) {
 				g_debug ("Could not open scan_elements file '%s'", filename);
-				g_free (filename);
 				continue;
 			}
 			if (fscanf (sysfsfp, "%d", &ret) != 1) {
-				fclose (sysfsfp);
 				g_debug ("Could not read from scan_elements file '%s'", filename);
-				g_free (filename);
 				continue;
 			}
-			fclose (sysfsfp);
-			g_free (filename);
+			g_clear_pointer (&sysfsfp, fclose);
 
 			current = g_new0 (iio_channel_info, 1);
 
@@ -281,15 +271,13 @@ build_channel_array (const char        *device_dir,
 			}
 
 			index_name = g_strdup_printf ("%s_index", current->name);
+			g_clear_pointer (&filename, g_free);
 			filename = g_build_filename (scan_el_dir, index_name, NULL);
-			g_free (index_name);
 
 			sysfsfp = fopen (filename, "r");
 			if (sysfsfp == NULL)
 				goto error;
 			ret = fscanf (sysfsfp, "%u", &current->index);
-			fclose (sysfsfp);
-			g_free (filename);
 			if (ret != 1)
 				goto error;
 
@@ -301,6 +289,8 @@ build_channel_array (const char        *device_dir,
 							current->generic_name);
 			if ((ret < 0) && (ret != -ENOENT))
 				goto error;
+			if (current->scale == 0.0)
+				goto error;
 
 			ret = iioutils_get_param_float (&current->offset,
 							"offset",
@@ -324,7 +314,7 @@ build_channel_array (const char        *device_dir,
 				g_warning ("Could not parse name %s, generic name %s",
 					   current->name, current->generic_name);
 			} else {
-				g_ptr_array_add (array, current);
+				g_ptr_array_add (array, g_steal_pointer (&current));
 			}
 		}
 	}
@@ -361,7 +351,7 @@ _write_sysfs_int (const char *filename,
 		  int         val2)
 {
 	int ret = 0;
-	FILE *sysfsfp;
+	g_autoptr(FILE) sysfsfp = NULL;
 	int test;
 	char *temp;
 	temp = g_build_filename (basedir, filename, NULL);
@@ -375,8 +365,8 @@ _write_sysfs_int (const char *filename,
 		fprintf(sysfsfp, "%d %d", val, val2);
 	else
 		fprintf(sysfsfp, "%d", val);
+	g_clear_pointer (&sysfsfp, fclose);
 
-	fclose(sysfsfp);
 	if (verify) {
 		sysfsfp = fopen(temp, "r");
 		if (sysfsfp == NULL) {
@@ -390,7 +380,6 @@ _write_sysfs_int (const char *filename,
 				   val, temp);
 			ret = -1;
 		}
-		fclose(sysfsfp);
 	}
 error_free:
 	g_free (temp);
@@ -411,39 +400,29 @@ _write_sysfs_string (const char *filename,
 		     const char *val,
 		     int         verify)
 {
-	int ret = 0;
-	FILE *sysfsfp;
-	char *temp;
+	g_autoptr(FILE) sysfsfp = NULL;
+	g_autofree char *temp = NULL;
+	g_autofree char *contents = NULL;
 
 	temp = g_build_filename (basedir, filename, NULL);
 	sysfsfp = fopen (temp, "w");
 	if (sysfsfp == NULL) {
-		ret = -errno;
-		goto error_free;
+		return -errno;
 	}
 	fprintf(sysfsfp, "%s", val);
-	fclose(sysfsfp);
 
 	/* Verify? */
 	if (!verify)
-		goto error_free;
-	sysfsfp = fopen(temp, "r");
-	if (sysfsfp == NULL) {
-		ret = -errno;
-		goto error_free;
+		return 0;
+	g_clear_pointer (&sysfsfp, fclose);
+	if (!g_file_get_contents (temp, &contents, NULL, NULL) ||
+	    g_strcmp0 (g_strchomp (contents), val) != 0) {
+		g_warning ("Possible failure in string write of '%s' Should be '%s' written to %s\n",
+			   contents ?: "(null)", val, temp);
+		return -1;
 	}
-	if (fscanf(sysfsfp, "%s", temp) != 1 ||
-	    strcmp(temp, val) != 0) {
-		g_warning ("Possible failure in string write of %s Should be %s written to %s\\%s\n",
-			   temp, val, basedir, filename);
-		ret = -1;
-	}
-	fclose(sysfsfp);
 
-error_free:
-	g_free(temp);
-
-	return ret;
+	return 0;
 }
 
 /**
@@ -749,6 +728,7 @@ buffer_drv_data_free (BufferDrvData *buffer_data)
 	for (i = 0; i < buffer_data->channels_count; i++)
 		channel_info_free (buffer_data->channels[i]);
 	g_free (buffer_data->channels);
+	g_free (buffer_data);
 }
 
 BufferDrvData *
@@ -774,3 +754,23 @@ buffer_drv_data_new (GUdevDevice *device,
 	return buffer_data;
 }
 
+IIOSensorData *
+iio_sensor_data_new (gsize size)
+{
+	IIOSensorData *data;
+
+	g_return_val_if_fail (size != 0, NULL);
+
+	data = g_new0 (IIOSensorData, 1);
+	data->data = g_malloc0 (size);
+	return data;
+}
+
+void
+iio_sensor_data_free (IIOSensorData *data)
+{
+	if (data == NULL)
+		return;
+	g_free (data->data);
+	g_free (data);
+}
