@@ -7,8 +7,9 @@
  *
  */
 
-#define _GNU_SOURCE
+#include "config.h"
 
+#include <locale.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -18,6 +19,7 @@
 
 #include <gio/gio.h>
 #include <gudev/gudev.h>
+#include <polkit/polkit.h>
 #include "drivers.h"
 #include "orientation.h"
 
@@ -39,9 +41,12 @@ typedef struct {
 	guint name_id;
 	int ret;
 
-	SensorDriver *drivers[NUM_SENSOR_TYPES];
-	GUdevDevice  *devices[NUM_SENSOR_TYPES];
-	GHashTable   *clients[NUM_SENSOR_TYPES]; /* key = D-Bus name, value = watch ID */
+	PolkitAuthority *auth;
+
+	SensorDriver      *drivers[NUM_SENSOR_TYPES];
+	SensorDevice      *devices[NUM_SENSOR_TYPES];
+	GUdevDevice       *udev_devices[NUM_SENSOR_TYPES];
+	GHashTable        *clients[NUM_SENSOR_TYPES]; /* key = D-Bus name, value = watch ID */
 
 	/* Accelerometer */
 	OrientationUp previous_orientation;
@@ -61,8 +66,8 @@ static const SensorDriver * const drivers[] = {
 	&iio_buffer_accel,
 	&iio_poll_accel,
 	&input_accel,
-	&iio_poll_light,
 	&iio_buffer_light,
+	&iio_poll_light,
 	&hwmon_light,
 	&fake_compass,
 	&fake_light,
@@ -91,6 +96,7 @@ driver_type_to_str (DriverType type)
 
 #define DRIVER_FOR_TYPE(driver_type) data->drivers[driver_type]
 #define DEVICE_FOR_TYPE(driver_type) data->devices[driver_type]
+#define UDEV_DEVICE_FOR_TYPE(driver_type) data->udev_devices[driver_type]
 
 static void sensor_changes (GUdevClient *client,
 			    gchar       *action,
@@ -124,13 +130,13 @@ find_sensors (GUdevClient *client,
 
 		for (i = 0; i < G_N_ELEMENTS(drivers); i++) {
 			SensorDriver *driver = (SensorDriver *) drivers[i];
-			if (!driver_type_exists(data, driver->type) &&
+			if (!driver_type_exists (data, driver->type) &&
 			    driver_discover (driver, dev)) {
 				g_debug ("Found device %s of type %s at %s",
 					 g_udev_device_get_sysfs_path (dev),
 					 driver_type_to_str (driver->type),
-					 driver->name);
-				DEVICE_FOR_TYPE(driver->type) = g_object_ref (dev);
+					 driver->driver_name);
+				UDEV_DEVICE_FOR_TYPE(driver->type) = g_object_ref (dev);
 				DRIVER_FOR_TYPE(driver->type) = (SensorDriver *) driver;
 
 				found = TRUE;
@@ -185,19 +191,36 @@ typedef enum {
 #define PROP_ALL_COMPASS (PROP_HAS_COMPASS | \
 			  PROP_COMPASS_HEADING)
 
+static PropertiesMask
+mask_for_sensor_type (DriverType sensor_type)
+{
+	switch (sensor_type) {
+	case DRIVER_TYPE_ACCEL:
+		return PROP_HAS_ACCELEROMETER |
+			PROP_ACCELEROMETER_ORIENTATION;
+	case DRIVER_TYPE_LIGHT:
+		return PROP_HAS_AMBIENT_LIGHT |
+			PROP_LIGHT_LEVEL;
+	case DRIVER_TYPE_COMPASS:
+		return PROP_HAS_COMPASS |
+			PROP_COMPASS_HEADING;
+	case DRIVER_TYPE_PROXIMITY:
+		return PROP_HAS_PROXIMITY |
+			PROP_PROXIMITY_NEAR;
+	default:
+		g_assert_not_reached ();
+	}
+}
+
 static void
-send_dbus_event (SensorData     *data,
-		 PropertiesMask  mask)
+send_dbus_event_for_client (SensorData     *data,
+			    const char     *destination_bus_name,
+			    PropertiesMask  mask)
 {
 	GVariantBuilder props_builder;
 	GVariant *props_changed = NULL;
 
-	g_assert (data->connection);
-
-	if (mask == 0)
-		return;
-
-	g_assert ((mask & PROP_ALL) == 0 || (mask & PROP_ALL_COMPASS) == 0);
+	g_return_if_fail (destination_bus_name != NULL);
 
 	g_variant_builder_init (&props_builder, G_VARIANT_TYPE ("a{sv}"));
 
@@ -278,13 +301,50 @@ send_dbus_event (SensorData     *data,
 				       g_variant_new_strv (NULL, 0));
 
 	g_dbus_connection_emit_signal (data->connection,
-				       NULL,
+				       destination_bus_name,
 				       (mask & PROP_ALL) ? SENSOR_PROXY_DBUS_PATH : SENSOR_PROXY_COMPASS_DBUS_PATH,
 				       "org.freedesktop.DBus.Properties",
 				       "PropertiesChanged",
 				       props_changed, NULL);
 }
 
+static void
+send_dbus_event (SensorData     *data,
+		 PropertiesMask  mask)
+{
+	GHashTable *ht;
+	guint i;
+	GHashTableIter iter;
+	gpointer key, value;
+
+	g_assert (mask != 0);
+	g_assert (data->connection);
+	g_assert ((mask & PROP_ALL) == 0 || (mask & PROP_ALL_COMPASS) == 0);
+
+	/* Make a list of the events each client for each sensor
+	 * is interested in */
+	ht = g_hash_table_new (g_str_hash, g_str_equal);
+	for (i = 0; i < NUM_SENSOR_TYPES; i++) {
+		GList *clients, *l;
+
+		clients = g_hash_table_get_keys (data->clients[i]);
+		for (l = clients; l != NULL; l = l->next) {
+			PropertiesMask m, new_mask;
+
+			/* Already have a mask? */
+			m = GPOINTER_TO_UINT (g_hash_table_lookup (ht, l->data));
+			new_mask = mask & mask_for_sensor_type (i);
+			m |= new_mask;
+			g_hash_table_insert (ht, l->data, GUINT_TO_POINTER (m));
+		}
+	}
+
+	g_hash_table_iter_init (&iter, ht);
+	while (g_hash_table_iter_next (&iter, &key, &value))
+		send_dbus_event_for_client (data, (const char *) key, GPOINTER_TO_UINT (value));
+	g_hash_table_destroy (ht);
+}
+
 static void
 send_driver_changed_dbus_event (SensorData   *data,
 				DriverType    driver_type)
@@ -328,14 +388,18 @@ client_release (SensorData            *data,
 	ht = data->clients[driver_type];
 
 	watch_id = GPOINTER_TO_UINT (g_hash_table_lookup (ht, sender));
-	if (watch_id == 0)
+	if (watch_id == 0) {
+		g_debug ("Sender '%s' already released device, no-op", sender);
 		return;
+	}
 
 	g_hash_table_remove (ht, sender);
 
 	if (driver_type_exists (data, driver_type) &&
-	    g_hash_table_size (ht) == 0)
-		driver_set_polling (DRIVER_FOR_TYPE(driver_type), FALSE);
+	    g_hash_table_size (ht) == 0) {
+		SensorDevice *sensor_device = DEVICE_FOR_TYPE(driver_type);
+		driver_set_polling (sensor_device, FALSE);
+	}
 }
 
 static void
@@ -367,6 +431,34 @@ client_vanished_cb (GDBusConnection *connection,
 	g_free (sender);
 }
 
+static gboolean
+check_claim_permission (SensorData   *data,
+			const char   *sender,
+			GError      **error)
+{
+	g_autoptr(GError) local_error = NULL;
+	g_autoptr(PolkitAuthorizationResult) result = NULL;
+	g_autoptr(PolkitSubject) subject = NULL;
+
+	subject = polkit_system_bus_name_new (sender);
+	result = polkit_authority_check_authorization_sync (data->auth,
+							    subject,
+							    "net.hadess.SensorProxy.claim-sensor",
+							    NULL,
+							    POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE,
+							    NULL, &local_error);
+	if (result == NULL ||
+	    !polkit_authorization_result_get_is_authorized (result))
+	{
+		g_set_error (error, G_DBUS_ERROR,
+			     G_DBUS_ERROR_ACCESS_DENIED,
+			     "Not Authorized: %s", local_error ? local_error->message : "Sensor claim not allowed");
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 static void
 handle_generic_method_call (SensorData            *data,
 			    const gchar           *sender,
@@ -388,14 +480,17 @@ handle_generic_method_call (SensorData            *data,
 	if (g_str_has_prefix (method_name, "Claim")) {
 		watch_id = GPOINTER_TO_UINT (g_hash_table_lookup (ht, sender));
 		if (watch_id > 0) {
+			g_debug ("Sender '%s' already claimed device, no-op", sender);
 			g_dbus_method_invocation_return_value (invocation, NULL);
 			return;
 		}
 
 		/* No other clients for this sensor? Start it */
 		if (driver_type_exists (data, driver_type) &&
-		    g_hash_table_size (ht) == 0)
-			driver_set_polling (DRIVER_FOR_TYPE(driver_type), TRUE);
+		    g_hash_table_size (ht) == 0) {
+			SensorDevice *sensor_device = DEVICE_FOR_TYPE(driver_type);
+			driver_set_polling (sensor_device, TRUE);
+		}
 
 		watch_id = g_bus_watch_name_on_connection (data->connection,
 							   sender,
@@ -425,6 +520,7 @@ handle_method_call (GDBusConnection       *connection,
 {
 	SensorData *data = user_data;
 	DriverType driver_type;
+	g_autoptr(GError) error = NULL;
 
 	if (g_strcmp0 (method_name, "ClaimAccelerometer") == 0 ||
 	    g_strcmp0 (method_name, "ReleaseAccelerometer") == 0)
@@ -444,6 +540,11 @@ handle_method_call (GDBusConnection       *connection,
 		return;
 	}
 
+	if (!check_claim_permission (data, sender, &error)) {
+		g_dbus_method_invocation_return_gerror (invocation, error);
+		return;
+	}
+
 	handle_generic_method_call (data, sender, object_path,
 				    interface_name, method_name,
 				    parameters, invocation, driver_type);
@@ -594,16 +695,22 @@ name_acquired_handler (GDBusConnection *connection,
 			  G_CALLBACK (sensor_changes), data);
 
 	for (i = 0; i < NUM_SENSOR_TYPES; i++) {
+		SensorDevice *sensor_device;
+
 		data->clients[i] = create_clients_hash_table ();
 
 		if (!driver_type_exists (data, i))
 			continue;
 
-		if (!driver_open (DRIVER_FOR_TYPE(i), DEVICE_FOR_TYPE(i),
-				  driver_type_to_callback_func (data->drivers[i]->type), data)) {
+		sensor_device = driver_open (DRIVER_FOR_TYPE(i), UDEV_DEVICE_FOR_TYPE(i),
+					     driver_type_to_callback_func (data->drivers[i]->type), data);
+		if (!sensor_device) {
 			DRIVER_FOR_TYPE(i) = NULL;
-			g_clear_object (&DEVICE_FOR_TYPE(i));
+			g_clear_object (&UDEV_DEVICE_FOR_TYPE(i));
+			continue;
 		}
+
+		DEVICE_FOR_TYPE(i) = sensor_device;
 	}
 
 	if (!any_sensors_left (data))
@@ -620,9 +727,11 @@ bail:
 }
 
 static gboolean
-setup_dbus (SensorData *data)
+setup_dbus (SensorData *data,
+	    gboolean    replace)
 {
 	GBytes *bytes;
+	GBusNameOwnerFlags flags;
 
 	bytes = g_resources_lookup_data ("/net/hadess/SensorProxy/net.hadess.SensorProxy.xml",
 					 G_RESOURCE_LOOKUP_FLAGS_NONE,
@@ -631,9 +740,13 @@ setup_dbus (SensorData *data)
 	g_bytes_unref (bytes);
 	g_assert (data->introspection_data != NULL);
 
+	flags = G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT;
+	if (replace)
+		flags |= G_BUS_NAME_OWNER_FLAGS_REPLACE;
+
 	data->name_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
 					SENSOR_PROXY_DBUS_NAME,
-					G_BUS_NAME_OWNER_FLAGS_NONE,
+					flags,
 					bus_acquired_handler,
 					name_acquired_handler,
 					name_lost_handler,
@@ -644,19 +757,22 @@ setup_dbus (SensorData *data)
 }
 
 static void
-accel_changed_func (SensorDriver *driver,
+accel_changed_func (SensorDevice *sensor_device,
 		    gpointer      readings_data,
 		    gpointer      user_data)
 {
 	SensorData *data = user_data;
 	AccelReadings *readings = (AccelReadings *) readings_data;
-	OrientationUp orientation = data->previous_orientation;
+	OrientationUp orientation;
 
 	//FIXME handle errors
-	g_debug ("Accel sent by driver (quirk applied): %d, %d, %d (scale: %lf)",
-		 readings->accel_x, readings->accel_y, readings->accel_z, readings->scale);
+	g_debug ("Accel sent by driver (quirk applied): %d, %d, %d (scale: %lf,%lf,%lf)",
+		 readings->accel_x, readings->accel_y, readings->accel_z,
+		 readings->scale.x, readings->scale.y, readings->scale.z);
 
-	orientation = orientation_calc (data->previous_orientation, readings->accel_x, readings->accel_y, readings->accel_z, readings->scale);
+	orientation = orientation_calc (data->previous_orientation,
+					readings->accel_x, readings->accel_y, readings->accel_z,
+					readings->scale);
 
 	if (data->previous_orientation != orientation) {
 		OrientationUp tmp;
@@ -671,7 +787,7 @@ accel_changed_func (SensorDriver *driver,
 }
 
 static void
-light_changed_func (SensorDriver *driver,
+light_changed_func (SensorDevice *sensor_device,
 		    gpointer      readings_data,
 		    gpointer      user_data)
 {
@@ -698,7 +814,7 @@ light_changed_func (SensorDriver *driver,
 }
 
 static void
-compass_changed_func (SensorDriver *driver,
+compass_changed_func (SensorDevice *sensor_device,
                       gpointer      readings_data,
                       gpointer      user_data)
 {
@@ -722,7 +838,7 @@ compass_changed_func (SensorDriver *driver,
 }
 
 static void
-proximity_changed_func (SensorDriver *driver,
+proximity_changed_func (SensorDevice *sensor_device,
 			gpointer      readings_data,
 			gpointer      user_data)
 {
@@ -779,11 +895,12 @@ free_sensor_data (SensorData *data)
 
 	for (i = 0; i < NUM_SENSOR_TYPES; i++) {
 		if (driver_type_exists (data, i))
-			driver_close (DRIVER_FOR_TYPE(i));
-		g_clear_object (&DEVICE_FOR_TYPE(i));
+			driver_close (DEVICE_FOR_TYPE(i));
+		g_clear_object (&UDEV_DEVICE_FOR_TYPE(i));
 		g_clear_pointer (&data->clients[i], g_hash_table_unref);
 	}
 
+	g_clear_object (&data->auth);
 	g_clear_pointer (&data->introspection_data, g_dbus_node_info_unref);
 	g_clear_object (&data->connection);
 	g_clear_object (&data->client);
@@ -804,7 +921,7 @@ sensor_changes (GUdevClient *client,
 
 	if (g_strcmp0 (action, "remove") == 0) {
 		for (i = 0; i < NUM_SENSOR_TYPES; i++) {
-			GUdevDevice *dev = DEVICE_FOR_TYPE(i);
+			GUdevDevice *dev = UDEV_DEVICE_FOR_TYPE(i);
 
 			if (!dev)
 				continue;
@@ -814,7 +931,9 @@ sensor_changes (GUdevClient *client,
 					 driver_type_to_str (i),
 					 g_udev_device_get_sysfs_path (dev));
 
-				g_clear_object (&DEVICE_FOR_TYPE(i));
+				g_clear_object (&UDEV_DEVICE_FOR_TYPE(i));
+				driver_close (DEVICE_FOR_TYPE(i));
+				DEVICE_FOR_TYPE(i) = NULL;
 				DRIVER_FOR_TYPE(i) = NULL;
 
 				g_clear_pointer (&data->clients[i], g_hash_table_unref);
@@ -834,20 +953,22 @@ sensor_changes (GUdevClient *client,
 				g_debug ("Found hotplugged device %s of type %s at %s",
 					 g_udev_device_get_sysfs_path (device),
 					 driver_type_to_str (driver->type),
-					 driver->name);
+					 driver->driver_name);
 
 				if (driver_open (driver, device,
 						 driver_type_to_callback_func (driver->type), data)) {
 					GHashTable *ht;
 
-					DEVICE_FOR_TYPE(driver->type) = g_object_ref (device);
+					UDEV_DEVICE_FOR_TYPE(driver->type) = g_object_ref (device);
 					DRIVER_FOR_TYPE(driver->type) = (SensorDriver *) driver;
 					send_driver_changed_dbus_event (data, driver->type);
 
 					ht = data->clients[driver->type];
 
-					if (g_hash_table_size (ht) > 0)
-						driver_set_polling (DRIVER_FOR_TYPE(driver->type), TRUE);
+					if (g_hash_table_size (ht) > 0) {
+						SensorDevice *sensor_device = DEVICE_FOR_TYPE(driver->type);
+						driver_set_polling (sensor_device, TRUE);
+					}
 				}
 				break;
 			}
@@ -858,15 +979,43 @@ sensor_changes (GUdevClient *client,
 int main (int argc, char **argv)
 {
 	SensorData *data;
+	g_autoptr(GOptionContext) option_context = NULL;
+	g_autoptr(GError) error = NULL;
+	gboolean verbose = FALSE;
+	gboolean replace = FALSE;
+	const GOptionEntry options[] = {
+		{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose, "Show extra debugging information", NULL },
+		{ "replace", 'r', 0, G_OPTION_ARG_NONE, &replace, "Replace the running instance of iio-sensor-proxy", NULL },
+		{ NULL}
+	};
 	int ret = 0;
 
+	setlocale (LC_ALL, "");
+	option_context = g_option_context_new ("");
+	g_option_context_add_main_entries (option_context, options, NULL);
+
+	ret = g_option_context_parse (option_context, &argc, &argv, &error);
+	if (!ret) {
+		g_print ("Failed to parse arguments: %s\n", error->message);
+		return EXIT_FAILURE;
+	}
+
+	if (verbose) {
+		if (!g_setenv ("G_MESSAGES_DEBUG", "all", TRUE)) {
+			g_warning ("Failed to enable debug");
+			return EXIT_FAILURE;
+		}
+		g_debug ("Starting iio-sensor-proxy version "VERSION);
+	}
+
 	data = g_new0 (SensorData, 1);
 	data->previous_orientation = ORIENTATION_UNDEFINED;
 	data->uses_lux = TRUE;
 
 	/* Set up D-Bus */
-	setup_dbus (data);
+	setup_dbus (data, replace);
 
+	data->auth = polkit_authority_get_sync (NULL, NULL);
 	data->loop = g_main_loop_new (NULL, TRUE);
 	g_main_loop_run (data->loop);
 	ret = data->ret;
